from typing import (
    Callable,
    Optional,
    Type,
    TypeVar,
    Union,
    get_type_hints,
    overload,
)

import functools
import inspect

from app_properties.config_handler import ConfigHandler
from app_properties.type_converter import TypeConverter

_T = TypeVar("_T")


@overload
def properties(
    cls: None = None,
    *,
    filename: str = ...,
    ignore_case: bool = ...,
    override_default: bool = ...,
    root: str = ...,
    type_cast: bool = ...,
) -> Callable[[Type[_T]], Type[_T]]:
    ...


@overload
def properties(
    cls: Type[_T],
    *,
    filename: str = ...,
    ignore_case: bool = ...,
    override_default: bool = ...,
    root: str = ...,
    type_cast: bool = ...,
) -> Type[_T]:
    ...


def properties(
    cls: Optional[Type[_T]] = None,
    *,
    filename: str = "application.yml",
    ignore_case: bool = True,
    override_default: bool = False,
    root: str = "",
    type_cast: bool = True,
) -> Union[Callable[[Type[_T]], Type[_T]], Type[_T]]:
    @functools.wraps(cls)  # type: ignore
    def wrapper(obj: Type[_T]) -> Type[_T]:
        type_converter = TypeConverter()
        config_handler = ConfigHandler(filename)
        annotated_class_vars = get_type_hints(obj)
        default_class_vars = _get_default_class_var(obj)
        config = config_handler.get_config(ignore_case, root)
        for class_var, type_ in annotated_class_vars.items():
            if not override_default and class_var in default_class_vars:
                continue
            value = config.get(class_var.lower() if ignore_case else class_var)
            if type_cast:
                value = type_converter.cast_types(type_, value)
            setattr(obj, class_var, value)
        return obj

    if cls is None:
        return wrapper
    return wrapper(cls)


def _get_default_class_var(obj: type) -> dict:
    return {
        k: v
        for k, v in inspect.getmembers(obj)
        if not (
            k.startswith("__")
            or callable(v)
            or isinstance(v, (property, classmethod))
        )
    }



from typing import Callable, Dict, TypeVar

import functools
import inspect
import json
import pathlib
import yaml
from os.path import dirname, join, normpath, sep

_K = TypeVar("_K")
_V = TypeVar("_V")
_T = TypeVar("_T")


class ConfigHandler:
    def __init__(self, config_name: str) -> None:
        self._conf_name = config_name
        self._caller_dir = self._get_caller_directory()

    def get_config(self, ignore_case: bool, root: str) -> dict:
        file = self._get_config_file()
        raw_config = self._resolve_config_format(file)
        return self._process_config(raw_config, ignore_case, root)

    def _get_caller_directory(self) -> str:
        stack = inspect.stack()
        stack_depth = 3
        # if used decorator without parens:
        if stack[stack_depth].filename.split(sep)[-2:] == [
            "app_properties",
            "main.py",
        ]:
            stack_depth = 4
        return dirname(stack[stack_depth].filename)

    def _resolve_config_format(self, file: pathlib.Path) -> dict:
        text_content = file.read_text()
        if file.suffix in (".yml", ".yaml"):
            conf = self._get_yaml_config(text_content)
        elif file.suffix == ".json":
            conf = self._get_json_config(text_content)
        else:
            raise NotImplementedError("Specified config type isn't supported!")
        return conf

    def _get_config_file(self) -> pathlib.Path:
        abs_config_path = join(self._caller_dir, normpath(self._conf_name))
        file = pathlib.Path(abs_config_path)
        if not file.exists():
            raise FileNotFoundError(f"File '{file.name}' is not found!")
        return file

    def _get_yaml_config(self, text_content: str) -> dict:
        return yaml.load(text_content, yaml.CSafeLoader)

    def _get_json_config(self, text_content: str) -> dict:
        return json.loads(text_content)

    def _apply_to_key(
        self, mapping: Dict[_K, _V], func: Callable[[_K], _T]
    ) -> Dict[_T, _V]:
        return {func(k): mapping[k] for k in mapping}

    def _process_config(
        self, config: dict, ignore_case: bool, root: str
    ) -> dict:
        config = self._apply_to_key(config, lambda x: str.replace(x, "-", "_"))
        if ignore_case:
            config = self._apply_to_key(config, str.lower)
        if root:
            config = functools.reduce(
                lambda x, y: x[y], root.split("."), config
            )
        return config



from typing import (
    Any,
    Optional,
    Tuple,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

import inspect
from collections.abc import Iterable, Mapping
from dataclasses import MISSING, Field, fields, is_dataclass
from itertools import zip_longest


class TypeConverter:
    def cast_types(self, type_: Union[type, Any], value: Any) -> Any:
        args = args if (args := get_args(type_)) else (Any,)
        type_ = origin if (origin := get_origin(type_)) else type_
        if (
            isinstance(type_, TypeVar)
            or type_ == Any
            or self._is_terminate(type_, args)
            and isinstance(value, type_)
        ):
            return value
        if type_ == Union or type_.__name__ == "UnionType":
            return self._apply_union(args, value)
        if issubclass(type_, list):
            return self._apply_list(args, value)
        if issubclass(type_, tuple):
            return self._apply_tuple(type_, args, value)
        if issubclass(type_, dict):
            return self._apply_dict(type_, args, value)
        if issubclass(type_, (set, frozenset)):
            return self._apply_set(type_, args, value)
        if is_dataclass(type_):
            return self._apply_dataclass(type_, value)
        return self._cast_base(type_, value)

    def _cast_base(self, type_: type, value: Any) -> Any:
        if value is None or self._is_none_type(type_):
            return type_()
        if type_ == bool:
            return self._cast_bool(value)
        return type_(value)

    def _cast_bool(self, value: Any) -> Optional[bool]:
        if isinstance(value, bool):
            return value
        value = str(value)
        if value in ("true", "True", "1"):
            return True
        if value in ("false", "False", "0"):
            return False
        return None

    def _apply_list(self, args: Tuple[type, ...], values: Any) -> list:
        if values is None:
            values = list()
        return [self.cast_types(args[0], i) for i in values]

    def _apply_set(
        self, type_: type, args: Tuple[type, ...], values: Any
    ) -> set:
        if values is None:
            values = set()
        return type_(self.cast_types(args[0], i) for i in values)

    def _apply_tuple(self, type_: type, args: Any, values: Any) -> tuple:
        if values is None:
            values = tuple()
        if not get_type_hints(type_):
            args = (Any, ...) if self._is_unspecified(args) else args
            args = self._resolve_ellipsis(args, values)
            cast_values: Any = [
                self.cast_types(arg, value) for arg, value in zip(args, values)
            ]
            return tuple(cast_values)
        if isinstance(values, Mapping):
            cast_values = {}
            for field, arg in get_type_hints(type_).items():
                cast_values[field] = self.cast_types(arg, values.get(field))
            return type_(**cast_values)
        if isinstance(values, Iterable):
            cast_values = []
            for arg, value in zip_longest(
                get_type_hints(type_).values(), values
            ):
                cast_values.append(self.cast_types(arg, value))
            return type_(*cast_values)
        raise ValueError("NamedTuple values should be iterable or mapping!")

    def _apply_dict(self, type_: type, args: tuple, values: Any) -> dict:
        if values is None:
            values = dict()
        if not isinstance(values, dict):
            raise ValueError("Value isn't mapping!")
        if not self._is_unspecified(args):
            return {
                self.cast_types(args[0], k): self.cast_types(args[1], v)
                for k, v in values.items()
            }
        cast_values = {}
        for field, arg in get_type_hints(type_).items():
            cast_values[field] = self.cast_types(arg, values.get(field))
        return type_(**cast_values)

    def _apply_dataclass(self, type_: type, values: Any) -> Any:
        if values is None:
            values = dict()
        field_mapping = {}
        for field in fields(type_):
            if not field.init:
                continue
            default = self._get_dataclass_default(field)
            field_mapping[field.name] = self.cast_types(
                field.type, values.get(field.name, default)
            )
        return type_(**field_mapping)

    def _get_dataclass_default(self, field: Field) -> Any:
        if field.default is not MISSING:
            return field.default
        if field.default_factory is not MISSING:
            return field.default_factory()
        return None

    def _resolve_ellipsis(
        self, args: Tuple[type, ...], values: Any
    ) -> Tuple[type, ...]:
        try:
            value_len = len(values)
        except TypeError:
            raise ValueError(f"Value {values} should be iterable!")
        if ... in args:
            return tuple(args[0] for _ in range(value_len))
        return args

    def _apply_union(self, args: Any, value: Any) -> Any:
        is_optional = type(None) in args
        if is_optional and value is None:
            return None
        args = [i for i in args if not self._is_none_type(i)]
        for arg in args:
            try:
                return self.cast_types(arg, value)
            except ValueError:
                continue
        if is_optional:
            return None
        raise ValueError(f"Couldn't cast '{value}' to any of types: {args}")

    def _is_terminate(self, type_: Any, args: tuple) -> bool:
        return (
            self._is_unspecified(args)
            and not self._is_primitive(type_)
            and not get_type_hints(type_)
        )

    def _is_primitive(self, type_: type) -> bool:
        return type_ in (str, int, float, bool, type(None))

    def _is_none_type(self, arg: Any) -> bool:
        return inspect.isclass(arg) and issubclass(arg, type(None))

    def _is_unspecified(self, args: Tuple[type, ...]) -> bool:
        return args == (Any,)



import io
import re
import setuptools

with io.open("app_properties/__init__.py", encoding="utf-8") as f:
    regex_result = re.search(r"__version__ = \"(.+)\"", f.read())
    version = regex_result.group(1) if regex_result else "0.1.0"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="py-app-properties",
    version=version,
    author="Yaroslav Kikvadze",
    author_email="yaroslav.k@simporter.com",
    description="Java application properties for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yar-kik/py-app-properties",
    license="MIT",
    packages=["app_properties"],
    install_requires=["PyYAML>=6.0"],
    python_requires=">=3.8",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)
